##Hokkaido University Research Center for Zoonosis Control
##Bioinformatics Department
##Gabriel Gonzalez, PhD. [mailto:gagonzalez@czc.hokudai.ac.jp
##2016
##########################################################
#This PERL script takes a variety of combinations of data of reads in fasta format, 
#	assuming fasta header line followed by the sequence line (2 lines per read).
#The first parameter is where to locate all the files to be processed, i.e. the working directory
#The second parameter is file with reads in fasta format to be used for the assembly
#The third parameter is a file in BLAST tab format that is used as the reference for the mapping of the different proteins
#	and used as guide to deduce the sequences assembling the reads from the first input
#The fourth parameter is the number of different proteins/sequences to assembly from all the possible entries in the tab file

#The results of this script are:
#	-Summary file showing the proteins in descending order by number of reads aligned to them
#	-A number of specified "contig" files deduced from the aligned reads is generated for the protein/sequences in summary or target protein

#COMMAND: perl align_reads_to_proteins_from_tab_results.pl [DIRECTORY] [FASTA file of reads] [BLAST_TAB file for reads] [max_number of contigs]
#	-DIRECTORY: the working directory where the reads and target file are located
#	-FASTA file of reads: fasta file of the reads to be aligned
#	-BLAST_TAB file with alignment info: used to guide the aligned positions to be considered matching the protein/sequence being assembled
#	-max_number of contigs: in case to be compared against a database, how many contigs will be attempted to assembly from the aligned reads

#!/usr/bin/perl
use feature ':5.10';
use strict;
use warnings;
use Data::Dumper qw(Dumper);

if($ARGV[0] eq "-h"|| $ARGV[0] eq "-help"){ #Option to provide a short information about how to execute the program
	die "perl align_reads_to_proteins_from_tab_results.pl [DIRECTORY] [FASTA file of reads] [BLAST_TAB file for reads] [max_number of contigs]";
}
#CONSTANTS
my $DIRECTORY_SEPARATOR = "\\"; 		#This depends on the platform, currently set for Unix, please change here in case of using it on Windows
my $FILE_EXTENSION = ".txt";			#Default file extension
my $KEY_SEPARATOR_PATTERN = "%%%%%%---%%%%%"; 	#hopefully no protein will include this pattern in their name in NCBI

my $WORKING_DIRECTORY = $ARGV[0];					#Directory with the read files to be processed
my $blast_results_file = $WORKING_DIRECTORY . $ARGV[2];	#Name of file with blast results
my $output_results_file = $WORKING_DIRECTORY . "Protein_";		#Suffix for deduced contigs
my $READS_FILE = $WORKING_DIRECTORY . $ARGV[1];				#File with the reads to be processed
my $PROTEIN_SUMMARY = $WORKING_DIRECTORY . "Aligned_Proteins.txt";	#Name of files with the aligned proteins sorted in descending order by the number of aligned reads
my $OUTPUT_LOG = $WORKING_DIRECTORY . "output.txt";			#Output log to show errors or messages
my $NUM_CONTIGS = $ARGV[3];						#Number of deduced contigs to be attempted
my $command = "";							#Variable to define the command to be executed, either BLAST or DIAMOND
my $adjust_aminoacid_nucleotide = 3; #factor to adjust the positions according to amino acids or nucleotides, since alignment is nucleotides

#remove spaces from the end and the beginning of the string
sub trim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	$string =~ s/\*//;
	return $string;
}

#remove the greater than sign of the fasta file to be consistent to the blastx results
sub fix_key($)
{
	my $string = shift;
	$string =~ s/^\s+//;
        $string =~ s/\s+$//;
        $string =~ s/\*//;
	$string =~ s/\s/_/g;
	$string =~ s/>//;
	return $string;
}

#returns the complement of a specific nucleotide
sub complement($)
{
	my $nt = shift;
	$nt = lc($nt);
	if($nt eq "a"){
		return "t";
	}
	if($nt eq "t"){
		return "a";
	}
	if($nt eq "g"){
		return "c";
	}
	if($nt eq "c"){
		return "g";
	}
	else{
		return "n";
	}
	die "Bad nucleotide " . $nt;
}

#Remove repeated elements from an array
sub uniq {
    my %seen;
    grep !$seen{$_}++, @_;
}

#EXECUTION:
#Step 1: using the alignment results and the read sequences a draft contig is assembled with the bases with highest support among the aligned reads. The reads
#are assembled in an optimistic approach: using the aligned region the rest of the read is considered even if it wasn't aligned to the protein
#Transform the blast results file into a hashtable for the aligned regions of each read. This hashtable requires to get only one hit per read
open READS_FILE, $READS_FILE or die "Error opening reads file " . $READS_FILE;
open BLASTX_RESULTS_FILE, $blast_results_file or die "Error opening blast results " . $blast_results_file;
my %blast_results;
my %aligned_proteins;

print("Reading Blast results\n");
while(<BLASTX_RESULTS_FILE>){		#Goes through all lines of the alignments results and transform the data into a hashtable indexed by the read and the aligned protein
	my $line = $_;
	my @data = split("\t", trim($line));
	if(defined($aligned_proteins{$data[1]})){
		$aligned_proteins{$data[1]} += 1;
	}else{
		$aligned_proteins{$data[1]} = 1;
	}
}
close(BLASTX_RESULTS_FILE);		#It closes the alignments results file

#Sort the proteins counts in descending order
open(OUTPUT_PROTEINS, '>', $PROTEIN_SUMMARY) or die 'Could not open file ' . $PROTEIN_SUMMARY . "\n";	#File to dump the aligned proteins sorted by number of reads
print OUTPUT_PROTEINS join("\t", "Descending.Ranking", "Protein", "#Reads") . "\n";			#Column names on the file
my $index = 0;												#Variable to account the rank of the sorting
my $count_targetted_proteins = 0;									#Counter to control the number of proteins for the number of contigs to deduce
my %targetted_proteins;											#Hash table of the proteins to be deduced

foreach my $protein (sort {$aligned_proteins{$b} <=> $aligned_proteins{$a}} keys %aligned_proteins) {	#Proteins are sorted according to the number of reads aligned to them
	print OUTPUT_PROTEINS join("\t", ++$index, $protein, $aligned_proteins{$protein}) . "\n";	#The protein is printed with the rank, name or identifier in the database and the count of aligned reads
	if($NUM_CONTIGS != 0){										#Only if further steps are required like deducing the contig
		$aligned_proteins{$protein} = $index;							#For the moment the table just keep the number of reads aligned to that protein
		if($count_targetted_proteins <= 100 && $count_targetted_proteins <= $NUM_CONTIGS){	#Controlling to account only proteins to be deduced in the contigs, if contigs are supposed to be deduced
			$targetted_proteins{$protein} = $index;						#The table provides reference later on to avoid deducing proteins that are out of scope
			$count_targetted_proteins++;							#The counter of considered proteins is increased
		}
	}
}

#Reading the results again to extract the reads related only with those proteins
#worth the effort
open BLASTX_RESULTS_FILE, $blast_results_file or die "Error opening blast results";
while(<BLASTX_RESULTS_FILE>){		#Goes through all lines of the alignments results and transform the data into a hashtable indexed by the read and the aligned protein
	my $line = $_;
	my @data = split("\t", trim($line));
	if(exists($targetted_proteins{$data[1]})){		#In case the protein is being targetted for deducing the contig
		$blast_results{'>' . $data[0]}{$data[1]} = [$data[6], $data[7], $adjust_aminoacid_nucleotide*($data[8]-1)+1, $adjust_aminoacid_nucleotide*($data[9]-1) + 1, $data[1]]; #There is an adjustment in the subject positions to translate from amino acid to codons
	}
}
close(BLASTX_RESULTS_FILE);		#It closes the alignments results file
#Another hashtable for all the base positions to be deduced for the aligned reads
my %deduced_positions;		#Otherwise it creates a hash table to start deducing the contigs

print("Constructing contigs from " . (keys %blast_results) . " aligned reads to " . (keys %aligned_proteins) . " proteins\n");

$index = 0;
my $sequence_number = 0;	#Counter for the number of read sequences processed
while(<READS_FILE>){		#The script goes through all reads to try to deduced the contigs from those aligned to the top proteins with the most sequences
	my $row_pos = trim($_);	#It reads the fasta header of the read
	$sequence_number += 1;	#It increases the counter of processed sequences
	if(($sequence_number % 10000) == 0){	#If it is a multiple of 10K it outputs the progress
		print("Processed reads: " . $sequence_number . " from the total\n");
	}
	if(exists($blast_results{$row_pos})){					#if the read was significantly aligned to any protein (according e-value parameter), the analysis can proceed
		my %proteins_alignments = %{$blast_results{$row_pos}};		#The lines of the blast results corresponding to the read are retrieve
		my $seq = <READS_FILE>;						#The fasta line is read from the reads file
		foreach my $new_key (keys %proteins_alignments){		#Iteration over the aligned proteins to that particular read
			if(exists($targetted_proteins{$new_key})){		#If the protein is among the proteins targetted among the top proteins to align ($NUM_CONTIGS)
				my $opposite_direction = $proteins_alignments{$new_key}[0] > $proteins_alignments{$new_key}[1];	#Testing if the direction of the alignment is the same than the protein alignment
				my $min_index = $proteins_alignments{$new_key}[0] <= $proteins_alignments{$new_key}[1] ? $proteins_alignments{$new_key}[0] : $proteins_alignments{$new_key}[1];	#Select what value is the left or right of the aligned protein
				for(my $i=0; $i < length($seq); $i++){		#Go through nucleotides in the read to deduce the contig
					my $char_seq = lc(substr($seq, $i, 1));	#Obtain the base in the position $i
					my $index_pos = ($proteins_alignments{$new_key}[2] <= $proteins_alignments{$new_key}[3] ? $proteins_alignments{$new_key}[2] : $proteins_alignments{$new_key}[3]) - $min_index + $i + 1; #deduce the position in the alignment
					if($opposite_direction){																				#Putting together the position of alignment and the base
						$char_seq = complement($char_seq);																		#Then correcting for complementarity of the bases
						$index_pos = ($proteins_alignments{$new_key}[2] >= $proteins_alignments{$new_key}[3] ? $proteins_alignments{$new_key}[2] : $proteins_alignments{$new_key}[3]) + $min_index - $i - 1; #And correcting the position accordingly
					}
					if(!($char_seq eq 'a') && !($char_seq eq 'c') && !($char_seq eq 'g') && !($char_seq eq 't')){			#Default case: undetermined base
						$char_seq = 'n';
					}
					if(!(exists($deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}))){	#If this position for that protein hasn't been deduced yet
						$deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}{'a'} = 0;	#Set the counts of adenine to 0
						$deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}{'c'} = 0;	#Set the counts of cytosine to 0
						$deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}{'g'} = 0;	#Set the counts of guanine to 0
						$deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}{'t'} = 0;	#Set the counts of thymine to 0
						$deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}{'n'} = 0;	#Set the counts of undetermined base to 0
					}
					my $value = $deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}{$char_seq} + 1;	#Extract the count so far for that position in that protein and adds 1
					$deduced_positions{$proteins_alignments{$new_key}[4] . $KEY_SEPARATOR_PATTERN . $index_pos}{$char_seq} = $value;	#set the current counter to the new value
				}
			}
		}
		delete($blast_results{$row_pos}); #Removes the entry from the hash table because the read was already processed and releases the memory
	}else{	#If there is no protein significantly aligned to that read, just move to the next read
		my $seq = <READS_FILE>;	#next line of the read should be ignored (sequence line in a FASTA file)
	}
}
close(READS_FILE); #It closes the reads file

print("Writing contigs\n");		#Writing the deduced contigs to files
my @contigs_proteins = sort {$aligned_proteins{$a} <=> $aligned_proteins{$b}} keys %aligned_proteins;	#Sort the aligned proteins into descending order
my @to_contigs = ();
for(my $i = 0; $i < @contigs_proteins && $i < $NUM_CONTIGS; $i++){
	push(@to_contigs, $contigs_proteins[$i]);
}
my @keys_deduced_positions = keys(%deduced_positions);	#Deduced proteins to show

foreach my $protein (@to_contigs) {
	my $pattern = $protein . $KEY_SEPARATOR_PATTERN;	#Building the pattern to extract the keys of the entries in the %deduced_positions related to that particular protein
	my @entries = grep {$_ =~ /\Q$pattern\E/} keys(%deduced_positions);	#extraction of keys related to the target protein
	my @positions = ();							#Array of positions
	foreach my $entry (@entries){						#Loop through entries related with the protein
		push(@positions, 0+(split($KEY_SEPARATOR_PATTERN, $entry))[1]);	#Extract the deduced positions of the sequence present in the hash table
	}
	@positions = sort {$a <=> $b} @positions;				#Sort positions in order
	open(OUTPUT, '>', $output_results_file . sprintf("%02d", $aligned_proteins{$protein}) . $FILE_EXTENSION) 
			or die 'Could not open file ' . $output_results_file . $aligned_proteins{$protein} . $FILE_EXTENSION . "\n";	#Create file for outputting the deduced contig
	print OUTPUT join("\t", "Protein", "Position", "a", "c", "g", "t", "n", "#Reads", "Base") . "\n";				#header for the file
	foreach my $pos (@positions){													#Loop through deduced positions to dump the deduced information
		my $sub_pos = $protein . $KEY_SEPARATOR_PATTERN . $pos;									#Key for the position
		my $a_count = exists($deduced_positions{$sub_pos}{'a'}) ? $deduced_positions{$sub_pos}{'a'} : 0;			#information of reads with 	a for that position
		my $c_count = exists($deduced_positions{$sub_pos}{'c'}) ? $deduced_positions{$sub_pos}{'c'} : 0;			#				g for that position
		my $g_count = exists($deduced_positions{$sub_pos}{'g'}) ? $deduced_positions{$sub_pos}{'g'} : 0;			#				c for that position
		my $t_count = exists($deduced_positions{$sub_pos}{'t'}) ? $deduced_positions{$sub_pos}{'t'} : 0;			#				t for that position
		my $n_count = exists($deduced_positions{$sub_pos}{'n'}) ? $deduced_positions{$sub_pos}{'n'} : 0;			#				n for that position
		my $number_reads =  $a_count + $c_count + $g_count + $t_count + $n_count;						#Total of reads aligned to this positions
		if($number_reads > 0){													#When the number of reads is greater than 0
			my @sorted = reverse(sort { $deduced_positions{$sub_pos}{$a} <=> $deduced_positions{$sub_pos}{$b} } keys %{$deduced_positions{$sub_pos}});	#the bases are sorted by count of supporting reads
			my $base = $sorted[0];												#The base is taken by the base with the highest count
			print OUTPUT join("\t", $protein,	
				$pos, $a_count, $c_count,$g_count, $t_count, $n_count, $number_reads, $base) . "\n";			#Line is printed to the output file
		}
	}
	close(OUTPUT);															#Close the output file
}

